Small Internet Protocol Stack using a standard serial port.

Dependencies:   mbed

PPP-Blinky - TCP/IP Networking Over a Serial Port

Note: The source code is at the bottom of this page.

/media/uploads/nixnax/blinky-connected.gif
A Windows desktop showing PPP-Blinky in the network connections list.

Describe PPP-Blinky in Three Sentences

PPP-Blinky is a tiny library that enables Internet protocols (IPv4) to any mbed target hardware by using only a serial port.

The code runs on processors with as little as 8k RAM, for example the Nucleo-L053R8 board.

PPP-Blinky uses the industry-standard PPP (Point-to-Point) Protocol and a tiny "stateless" TCP/IP stack.

No Ethernet Port Required

No ethernet port is required - PPP-Blinky uses a serial port to send IP packets to your PC.

PPP-Blinky emulates a standard dial-up modem and therefore connects to Windows, Linux or Adroid machines.

The code runs on most ARM mbed platforms such as the LPC11U24 shown in the picture below:

/media/uploads/nixnax/blinky-to-laptop1.jpg mbed LPC11u24 acting as a webserver to a Windows laptop.

Webserver

The Webserver and WebSocket functions are ideal for building browser-based GUIs on mbed-enabled hardware.

PPP-Blinky's HTTP webserver works with most web clients such as Internet Explorer, Mozilla Firefox, Google Chrome, Safari, Curl, wget and Lynx as well as Microsoft Powershell Invoke-Webrequest command.

In the image below Firefox web browser displays the main web page embedded into PPP-Blinky's code:

/media/uploads/nixnax/ppp-blinky-firefox.jpg Firefox web browser displays a web page embedded into PPP-Blinky's code

WebSocket Service

WebSocket is the most popular protocol standard for real-time bidirectional TCP/IP communication between clients and servers.
In the image below a small Internet Explorer script has connected to PPP-Blinky's WebSocket Service.
A websocket message was then sent by the browser and was echoed back by the WebSocket, triggering the onmessage event in the script.
The WebSocket service enables bidirectional real-time interaction between PPP-Blinky and any element in the browser DOM via JavaScript.
If you already have PPP-Blinky up and running you can test your WebSocket service using this: http://jsfiddle.net/d26cyuh2/112/embedded/result
Websockets are ideal for building browser-based GUIs for mbed hardware.

/media/uploads/nixnax/ppp-blinky-websocke-2.gif

Trying PPP-Blinky on your mbed board

You will need an mbed-enabled hardware board: https://developer.mbed.org/platforms/

Establish a serial port connection between your host PC and your mbed board. The easiest way is to use mbed hardware with a USB serial debug port. I've tried the ST-Micro Nucleo-L476RG, Nucleo-L152RE, Nucleo-F401RE, Nucleo-L432KC, Nucleo-L053R8, mbed-LPC11U24 and mbed-LPC1768 boards and they all work out of the box. Use the mbed online compiler to compile the software for your target board. Save the compiled binary to your hardware.

Before establishing a network connection, you can verify the operation of the code by opening a terminal program such as Tera Term, and setting the baud rate of the COM port on your mbed board to 115200 baud. LED1 should toggle for every two 0x7E (~) (i.e. tilde) characters you type, as 0x7E is the PPP frame start/end marker. Don't forget to close the port when your'e done testing, or else Windows Dial-up Networking will report that the COM port is in use by another program when you try to connect.

Once you are certain that the serial port and firmware is working, proceed to creating a new network connection on your PC -see below.

Creating a Dial-up Connection in Windows

/media/uploads/nixnax/modem.jpg

Setting up Dial-Up Networking (DUN) on your Windows 7 or 8 PC is essentially a two-step process: First, you create a new modem device, because PPP-blinky partially emulates a standard Windows serial port modem device. Second, you create a new Internet connection (in practice, a new network adapter) which is associated with your new "modem".

Step-by-step description of how to configure Windows for PPP-Blinky here:

/users/nixnax/code/PPP-Blinky/wiki/Configuring-Windows-Dial-Up-Networking

There is also a screen on how to set up Linux dial-up networking near the bottom of this page.

Connecting to PPP-Blinky from your PC

Once Windows networking is configured you can establish a dial-up connection to your mbed board over the USB virtual com port.

The IP address you manually assigned to the new dial-up network adapter (172.10.10.1) functions as a gateway to any valid IP address on that subnet. In the screen capture below, I'm sending pings from the Windows 8 command line to my ST-Micro Nucleo-L476RG board over the USB virtual serial Port. I'm also using a second serial port and Tera Term to capture the debug output from a second serial port on the hardware. The optional debug output from the board prints out the IP source and destination address and the first few bytes of the data payload. Note that the source is the adapter IP address, (172.10.10.1 in this case) and the destination is some other address on that subnet - all packets to the subnet are sent to our mbed hardware. For example, you could also ping 172.10.10.123 or, if your PPP-Blinky is running, simply click on this link: http://172.10.10.123

/media/uploads/nixnax/ping-cap-3.gif

One Million Pings!

In the image below the ICMP ("ping") echo reply service was tested by sending one million pings to ppp-Blinky. This took over two hours.
The ping tool used on the Windows 8 PC was psping.exe from PsTools by Mark Russinovich - http://bit.ly/PingFast
The average reply time for a short ping (1 byte of payload data) was 11 milliseconds at 115200 baud on the $10 Nucleo-L053R8 board - barely enough time for 130 bytes to be sent over the port!

/media/uploads/nixnax/ppp-blinky-ping-results.jpg

Monitoring PPP-Blinky Packets

The image below is from a Microsoft Network Monitor 3.4 capture session.

Responses from PPP-Blinky are shown in blue.

Frame 2 - Internet Explorer at IP 172.10.10.1 (the Dial-Up Adapter IP) requests a TCP connection by sending an S (SYN) flag.
Frame 3 - PPP-Blinky at IP 172.10.10.2 responds with an ACK in frame 3. One direction of the link is now established.
Frame 4 - The PC acknowledges the SYN sent by PPP-Blinky in frame 3. The TCP link is now fully established.
Frame 5 - The browser "pushes" (P flag is set) an HTTP GET request to PPP-Blinky.
Frame 6 - PPP-Blinky responds with a standard HTTP response "pushes" (P flag set) back a small web page. It also sets the A (ACK) flag to acknowledge the message sent in frame 6.
Frame 7 - The PC acknowledges reception of the HTTP payload.
Frame 8 - The PC starts to shut down the TCP connection by sending a FIN flag.
Frame 9 - PPP-Blinky acknowledges the FIN request - the connection is now closed in one direction. It also sets a FIN flag in the response to request closure of the opposite direction of the connection.
Frame 10 - The PC acknowledges the FIN request. The closing of the TCP connection is now confirmed in both directions.

/media/uploads/nixnax/ms-network-monitor-http-get-1.gif

Debug Output

PPP-Blinky can output handy debug information to an optional second serial port.
The image below shows the debug output (Ident, Source, Destination, TCP Flags) for a complete HTTP conversation.
The PC messages are displayed in black. PPP-Blinky messages are blue.
Notice how PPP-blinky automatically inserts a blank line after each full HTTP conversation.

/media/uploads/nixnax/tcp-data-3.gif

Creating a Dial-Up Connection in Linux

The screen below shows the required pppd command to connect to PPP-Blinky from a Linux machine. This was much simpler than Windows! The USB serial port of the mbed LPC1768 board registered as /dev/ttyACM0 on my Linux box. Do a websearch on pppd if you want to learn more about pppd, the Linux PPP handler. Near the bottom of the screen below, two webpages are fetched (/ and /y) by using the curl command on the command line. Gnome Webkit and Firefox work fine, too. Also try echo GET / HTTP/1.1 | nc 172.10.10.2 which uses netcat, the "Swiss army knife" of networking tools. PPP-Blinky was also tested with ApacheBench, the Apache server benchmark software. After 100000 fetches, the mean page fetch rate was reported as 6 page fetches per second for a small page.

/media/uploads/nixnax/pppd-screen.png

Caveats

PPP Blinky is an extremely sparse implementation (1.5k lines) of HTTP,WebSocket,TCP, UDP, ICMP, IPCP and LCP over PPP, requiring around 8kB of RAM. The minimum functionality required to establish connectivity is implemented. These are often acceptable tradeoffs for embedded projects as well as a handy tool to learn the practical details of everyday networking implementations.

Revision:
124:18ef53f1d8b7
Parent:
123:fc64fc6caae0
Child:
125:ea88200b1df6
--- a/main.cpp	Tue Aug 08 20:22:19 2017 +0000
+++ b/main.cpp	Wed Aug 09 12:37:01 2017 +0000
@@ -65,7 +65,7 @@
 
 // verbosity flags used in debug printouts - change to 1 to see increasingly more detailed debug info.
 #define v0 1
-#define v1 1
+#define v1 0
 #define v2 0
 #define IP_HEADER_DUMP_YES /* YES for ip header dump */
 #define TCP_HEADER_DUMP_YES /* YES for tcp header dump */
@@ -110,7 +110,7 @@
     int httpPageCount;
     int firstFrame; // cleared after first frame
     struct {
-#define RXBUFLEN (1<<11)
+#define RXBUFLEN (1<<15)
         // the serial port receive buffer and packet buffer, size is RXBUFLEN (currently 2048 bytes)
         char buf[RXBUFLEN]; // RXBUFLEN MUST be a power of two because we use & operator for fast wrap-around in ring buffer
         int head;
@@ -198,10 +198,16 @@
     while( *nextChar != 0 ) {
         putcWhileCheckingInput( *nextChar ); // write one character to debug port while checking input
         nextChar++;
-    }
+    } 
 #endif
 }
 
+// a sniffer tool to assist in figuring out where in the code we are having characters in the input buffer
+void qq()
+{
+    if ( pc.readable() ) printWhileCheckingInput( "Character available!\n" );
+}
+
 void crcReset()
 {
     ppp.crc=0xffff;   // crc restart
@@ -213,7 +219,7 @@
         ppp.crc=((ppp.crc&1)^(x&1))?(ppp.crc>>1)^0x8408:ppp.crc>>1; // crc calculator
         x>>=1;
     }
-    // fillbuf();
+    fillbuf(); // handle input
 }
 
 int crcBuf(char * buf, int size) // crc on an entire block of memory
@@ -223,12 +229,6 @@
     return ppp.crc;
 }
 
-int rxbufNotEmpty()   // check if rx buffer has data
-{
-    int emptyStatus = (ppp.rx.head==ppp.rx.tail) ? 0 : 1 ;
-    return emptyStatus;
-}
-
 int pc_getBuf()   // get one character from the buffer
 {
     int x = ppp.rx.buf[ ppp.rx.tail ];
@@ -244,15 +244,15 @@
 {
     char pbuf[30];
     for(int i=0; i<ppp.pkt.len; i++) {
-         fillbuf();
-         sprintf(pbuf, "%02x ", ppp.pkt.buf[i]);
-         fillbuf();
-         printWhileCheckingInput(pbuf);
-         }
-         fillbuf();
-         sprintf(pbuf, " CRC=%04x Len=%d\n", ppp.pkt.crc, ppp.pkt.len);
-         fillbuf();
-         printWhileCheckingInput(pbuf);
+        fillbuf();
+        sprintf(pbuf, "%02x ", ppp.pkt.buf[i]);
+        fillbuf();
+        printWhileCheckingInput(pbuf);
+    }
+    fillbuf();
+    sprintf(pbuf, " CRC=%04x Len=%d\n", ppp.pkt.crc, ppp.pkt.len);
+    fillbuf();
+    printWhileCheckingInput(pbuf);
 }
 
 void processPPPFrame(int start, int end)   // process received frame
@@ -267,6 +267,7 @@
     int unstuff=0;
     int idx = start;
     while(1) {
+        fillbuf();
         if (unstuff==0) {
             if (ppp.rx.buf[idx]==0x7d) unstuff=1;
             else {
@@ -291,27 +292,31 @@
         void determinePacketType(); // declaration only
         determinePacketType();
     } else {
-        if (0) { // set to 1 to report FCS errors
+#ifdef REPORT_FCS_ERROR_YES            
             char pbuf[50]; // local print buffer
             fillbuf();
-            sprintf(pbuf, "PPP FCS(crc) Error CRC=%x Length = %d\n",ppp.pkt.crc,ppp.pkt.len); // print a debug line
+            sprintf(pbuf, "\nPPP FCS(crc) Error CRC=%x Length = %d\n",ppp.pkt.crc,ppp.pkt.len); // print a debug line
             fillbuf();
             printWhileCheckingInput( pbuf );
             if(0) dumpPPPFrame(); // set to 1 to dump frames with errors in them
-        }
+#endif        
     }
 }
 
+void pcPutcWhileCheckingInput(int ch)
+{
+    fillbuf(); // check input
+    pc.putc(ch);
+    fillbuf();
+}
+
 void hdlcPut(int ch)   // do hdlc handling of special (flag) characters
 {
     if ( (ch<0x20) || (ch==0x7d) || (ch==0x7e) ) {
-        fillbuf();
-        pc.putc(0x7d);
-        fillbuf();
-        pc.putc(ch^0x20);  // these characters need special handling
+        pcPutcWhileCheckingInput(0x7d);
+        pcPutcWhileCheckingInput(ch^0x20);  // these characters need special handling
     } else {
-        fillbuf();
-        pc.putc(ch);
+        pcPutcWhileCheckingInput(ch);
     }
 }
 
@@ -320,11 +325,11 @@
     int crc = crcBuf(ppp.pkt.buf, ppp.pkt.len-2); // update crc
     ppp.pkt.buf[ ppp.pkt.len-2 ] = (~crc>>0); // fcs lo (crc)
     ppp.pkt.buf[ ppp.pkt.len-1 ] = (~crc>>8); // fcs hi (crc)
-    pc.putc(0x7e); // hdlc start-of-frame "flag"
+    pcPutcWhileCheckingInput(0x7e); // hdlc start-of-frame "flag"
     for(int i=0; i<ppp.pkt.len; i++) {
         hdlcPut( ppp.pkt.buf[i] ); // send a character
     }
-    pc.putc(0x7e); // hdlc end-of-frame "flag"
+    pcPutcWhileCheckingInput(0x7e); // hdlc end-of-frame "flag"
 }
 
 void ipcpConfigRequestHandler()
@@ -337,7 +342,7 @@
     ppp.pkt.buf[4]=1; // change code to request
     ppp.pkt.buf[7]=4; // no options in this request
     ppp.pkt.len=10; // no options in this request shortest ipcp packet possible (4 ppp + 4 ipcp + 2 crc)
-    send_pppFrame(); // send our request
+    send_pppFrame(); // send our request 
 }
 
 void ipcpAckHandler()
@@ -423,6 +428,7 @@
         ptr[len]=0;
     }
     for (int i=0; i<len/2; i++) {
+        fillbuf();
         unsigned int hi = *ptr;
         ptr++;
         unsigned int lo = *ptr;
@@ -451,6 +457,7 @@
         ptr++;
         int val = ( lo & 0xff ) | ( (hi<<8) & 0xff00 );
         sum = sum + val;
+        fillbuf();
     }
     sum = sum + (sum>>16);
     sum = ~sum;
@@ -459,7 +466,7 @@
 }
 
 void ICMPpacket()   // internet control message protocol
-{
+{ 
     char * ipPkt = ppp.pkt.buf+4; // ip packet start
     char * pktLen = ipPkt+2;
     int packetLength = (pktLen[0]<<8) | pktLen[1]; // icmp packet length
@@ -472,12 +479,20 @@
         ipTTL[0]--; // decrement time to live
         char * srcAdr = ipPkt+12;
         char * dstAdr = ipPkt+16;
-#ifndef SERIAL_PORT_MONITOR_NO
+#ifdef SERIAL_PORT_MONITOR_YES
         int icmpIdent = (icmpType[4]<<8)|icmpType[5];
         int icmpSequence = (icmpType[6]<<8)|icmpType[7];
-#endif
-        if(v0) debugPrintf("ICMP PING %d.%d.%d.%d %d.%d.%d.%d ", srcAdr[0],srcAdr[1],srcAdr[2],srcAdr[3],dstAdr[0],dstAdr[1],dstAdr[2],dstAdr[3]);
-        if(v0) debugPrintf("Ident %04x Sequence %04d ",icmpIdent,icmpSequence);
+        if(0) {
+            char pbuf[50];
+            fillbuf();
+            sprintf(pbuf, "ICMP PING %d.%d.%d.%d %d.%d.%d.%d ", srcAdr[0],srcAdr[1],srcAdr[2],srcAdr[3],dstAdr[0],dstAdr[1],dstAdr[2],dstAdr[3]);
+            printWhileCheckingInput( pbuf );
+            fillbuf();
+            sprintf(pbuf, "Ident %04x Sequence %04d ",icmpIdent,icmpSequence);
+            fillbuf();
+            printWhileCheckingInput( pbuf );
+        }
+#endif        
         char src[4];
         char dst[4];
         memcpy(src, srcAdr,4);
@@ -500,24 +515,23 @@
         int printSize = icmpLength-8; // exclude size of icmp header
         char * icmpData = icmpType+8; // the actual payload data is after the header
         if (printSize > 10) printSize = 10; // print up to 20 characters
-        if (v0) {
+        if (0) {
             for (int i=0; i<printSize; i++) {
                 char ch = icmpData[i];
                 if (ch>31 && ch<127) {
-                    debugPrintf("%c",ch);
+                    putcWhileCheckingInput(ch);
                 } else {
-                    debugPrintf("_");
+                    putcWhileCheckingInput('_');
                 }
             }
-            debugPrintf("\n");
+            putcWhileCheckingInput('\n');
         }
         send_pppFrame(); // reply to the ping
-
     } else {
         if (v0) {
             debugPrintf("ICMP type=%d \n", icmpType[0]);
         }
-    }
+    } 
 }
 
 void IGMPpacket()   // internet group management protocol
@@ -556,11 +570,8 @@
     int IPv4Id = (ident[0]<<8)|ident[1];
     char pbuf[50]; // local print buffer
     int n=0;
-    fillbuf(); // we are expecting the first character of the next packet
     n=n+sprintf(pbuf+n, outGoing ? "\x1b[34m" : "\x1b[30m" ); // VT100 color code, print black for incoming, blue for outgoing headers
-    fillbuf(); // we are expecting the first character of the next packet
     n=n+sprintf(pbuf+n, "%05d ",IPv4Id); // IPv4Id is a good way to correlate our dumps with net monitor or wireshark traces
-    fillbuf(); // we are expecting the first character of the next packet
 #define DUMP_FULL_IP_ADDRESS_YES
 #ifdef DUMP_FULL_IP_ADDRESS_YES
     char * srcAdr =     ipPkt+12; // 4 bytes
@@ -606,7 +617,7 @@
     if (flags & (1<<0)) flagInfo[0]='F';
     if (flags & (1<<3)) flagInfo[0]='P';
     if (flags & (1<<2)) flagInfo[0]='R';
-    flagInfo[1]=' ';
+    flagInfo[1]=0; // ' '
     flagInfo[2]=0;
 #endif
     printWhileCheckingInput( flagInfo );
@@ -724,6 +735,8 @@
     char * flagbitstcp   = tcp + 12; // 9 bits
     char * windowsizetcp = tcp + 14; // 2 bytes
     char * checksumtcp   = tcp + 16; // 2 bytes
+    
+    if(ident) {}; // shut up unused variable reference warning
 
     int tcpSize = packetLength - headerSizeIP;
     int headerSizeTCP = ((offset[0]>>4)&0x0f)*4; // size of tcp header only
@@ -814,8 +827,8 @@
 
     // increment our outgoing ip packet counter
     ppp.ip.ident++; // get next ident number for our packet
-    ident[0] = ppp.ip.ident>>8;
-    ident[1] = ppp.ip.ident>>0; // insert OUR ident
+    //ident[0] = ppp.ip.ident>>8;
+    //ident[1] = ppp.ip.ident>>0; // insert OUR ident
 
     // Now we recalculate all the header sizes
     int newPacketSize = headerSizeIP + headerSizeTCP + dataLen; // calculate size of the outgoing packet
@@ -849,9 +862,9 @@
     memcpy( tcp-12, tempHold, 12); // restore the 12 bytes that the pseudo-header overwrote
     dumpHeaderIP(1); // dump outgoing IP header
     dumpHeaderTCP(1); // dump outgoing TCP header
-    for (int i=0; i<60000/50; i++) { // a 60 ms delay before sending frame - typical internet delay time
+    for (int i=0; i<70*1000/10; i++) { // 70 ms delay before sending frame - a typical internet delay time
         fillbuf(); // catch any incoming characters
-        wait_us(50); // wait less than 1 character duration at 115200
+        wait_us(10); // wait less than 1 character duration at 115200
     }
     send_pppFrame(); // All preparation complete - send the TCP response
 }
@@ -877,12 +890,12 @@
 }
 
 void TCPpacket()
-{
+{ 
     dumpHeaderIP(0);     // dump incoming packet header
     dumpHeaderTCP(0);   // dump incoming packet header
     if (v2) {
         dumpDataTCP();
-    }
+    } 
     tcpHandler();
 }
 
@@ -1008,7 +1021,7 @@
 {
     while(1) {
         fillbuf(); // handle received characters
-        if ( rxbufNotEmpty() ) {
+        if ( ppp.rx.head != ppp.rx.tail ) {
             int oldTail = ppp.rx.tail; // remember where the character is located in the buffer
             int rx = pc_getBuf(); // get the character
             if (rx==FRAME_7E) {